from enum import Enum
from typing import Optional, List

__all__ = [
    'Alphabet',
    'AudioFormat',
    'ContentType',
    'Gender',
    'LanguageCode',
    'Region',
    'SpeechMarkTypes',
    'StrEnum',
    'SynthesisTaskStatus',
    'TextType',
    'VoiceID',
]


class StrEnum(str, Enum):
    def __str__(self):
        return self.value

    def __contains__(self, other):
        if isinstance(other, self.__class__):
            other = other.value
        return self.value == other

    @classmethod
    def search(cls, contain: str = None,
               startswith: str = None,
               endswith: str = None,
               contained: str = None) -> List['StrEnum']:
        """
        :param contain: search all enum members which are contain some substring
        :param startswith: search all enum members which are start with some substring
        :param endswith: search all enum members which are end with some substring
        :param contained: search all enum members which are substrings of some string

        :return: all found enums
        """

        found = set()
        if contain:
            found.update(e for e in cls if contain in e.value)
        if startswith:
            found.update(e for e in cls if e.value.startswith(startswith))
        if endswith:
            found.update(e for e in cls if e.value.endswith(endswith))
        if contained:
            found.update(e for e in cls if e.value in contained)

        return list(found)


class TextType(StrEnum):
    text = 'text'
    ssml = 'ssml'


class SpeechMarkTypes(StrEnum):
    ssml = 'ssml'
    sentence = 'sentence'
    viseme = 'viseme'
    word = 'word'


class SynthesisTaskStatus(StrEnum):
    scheduled = 'scheduled'
    in_progress = 'inProgress'
    completed = 'completed'
    failed = 'failed'


class Alphabet(StrEnum):
    ipa: str = 'ipa'
    x_sampa: str = 'x-sampa'
    x_amazon_pinyin: str = 'x-amazon-pinyin'


class Gender(Enum):
    female = 'Female'
    male = 'Male'


class AudioFormat(StrEnum):
    json = 'json'
    mp3 = 'mp3'
    pcm = 'pcm'
    ogg_vorbis = 'ogg_vorbis'


class ContentType(StrEnum):
    application_json = 'application/json'
    application_octet_stream = 'application/octet-stream'
    audio_json = 'application/x-json-stream'
    audio_mpeg = 'audio/mpeg'
    audio_pcm = 'audio/pcm'
    audio_ogg = 'audio/ogg'

    @property
    def audio_format(self) -> Optional[AudioFormat]:
        return AUDIO_CONTENT_TYPE_FORMAT_MAPPING.get(self)


class Region(StrEnum):
    us_east_2 = 'us-east-2'
    us_east_1 = 'us-east-1'
    us_west_1 = 'us-west-1'
    us_west_2 = 'us-west-2'
    ap_south_1 = 'ap-south-1'
    ap_northeast_2 = 'ap-northeast-2'
    ap_southeast_1 = 'ap-southeast-1'
    ap_southeast_2 = 'ap-southeast-2'
    ap_northeast_1 = 'ap-northeast-1'
    ca_central_1 = 'ca-central-1'
    cn_northwest_1 = 'cn-northwest-1'
    eu_central_1 = 'eu-central-1'
    eu_west_1 = 'eu-west-1'
    eu_west_2 = 'eu-west-2'
    eu_west_3 = 'eu-west-3'
    eu_north_1 = 'eu-north-1'
    sa_east_1 = 'sa-east-1'


class LanguageCode(StrEnum):
    arb = 'arb'
    cmn_CN = 'cmn-CN'
    cy_GB = 'cy-GB'
    da_DK = 'da-DK'
    de_DE = 'de-DE'
    en_AU = 'en-AU'
    en_GB = 'en-GB'
    en_GB_WLS = 'en-GB-WLS'
    en_IN = 'en-IN'
    en_US = 'en-US'
    es_ES = 'es-ES'
    es_MX = 'es-MX'
    es_US = 'es-US'
    fr_CA = 'fr-CA'
    fr_FR = 'fr-FR'
    is_IS = 'is-IS'
    it_IT = 'it-IT'
    ja_JP = 'ja-JP'
    hi_IN = 'hi-IN'
    ko_KR = 'ko-KR'
    nb_NO = 'nb-NO'
    nl_NL = 'nl-NL'
    pl_PL = 'pl-PL'
    pt_BR = 'pt-BR'
    pt_PT = 'pt-PT'
    ro_RO = 'ro-RO'
    ru_RU = 'ru-RU'
    sv_SE = 'sv-SE'
    tr_TR = 'tr-TR'


class VoiceID(StrEnum):
    Aditi = 'Aditi'
    Amy = 'Amy'
    Astrid = 'Astrid'
    Bianca = 'Bianca'
    Brian = 'Brian'
    Carla = 'Carla'
    Carmen = 'Carmen'
    Celine = 'Celine'
    Chantal = 'Chantal'
    Conchita = 'Conchita'
    Cristiano = 'Cristiano'
    Dora = 'Dora'
    Emma = 'Emma'
    Enrique = 'Enrique'
    Ewa = 'Ewa'
    Filiz = 'Filiz'
    Geraint = 'Geraint'
    Giorgio = 'Giorgio'
    Gwyneth = 'Gwyneth'
    Hans = 'Hans'
    Ines = 'Ines'
    Ivy = 'Ivy'
    Jacek = 'Jacek'
    Jan = 'Jan'
    Joanna = 'Joanna'
    Joey = 'Joey'
    Justin = 'Justin'
    Karl = 'Karl'
    Kendra = 'Kendra'
    Kimberly = 'Kimberly'
    Lea = 'Lea'
    Liv = 'Liv'
    Lotte = 'Lotte'
    Lucia = 'Lucia'
    Mads = 'Mads'
    Maja = 'Maja'
    Marlene = 'Marlene'
    Mathieu = 'Mathieu'
    Matthew = 'Matthew'
    Maxim = 'Maxim'
    Mia = 'Mia'
    Miguel = 'Miguel'
    Mizuki = 'Mizuki'
    Naja = 'Naja'
    Nicole = 'Nicole'
    Penelope = 'Penelope'
    Raveena = 'Raveena'
    Ricardo = 'Ricardo'
    Ruben = 'Ruben'
    Russell = 'Russell'
    Salli = 'Salli'
    Seoyeon = 'Seoyeon'
    Takumi = 'Takumi'
    Tatyana = 'Tatyana'
    Vicki = 'Vicki'
    Vitoria = 'Vitoria'
    Zeina = 'Zeina'
    Zhiyu = 'Zhiyu'


AUDIO_CONTENT_TYPES = (
    ContentType.audio_ogg,
    ContentType.audio_pcm,
    ContentType.audio_json,
    ContentType.audio_mpeg
)

AUDIO_CONTENT_TYPE_FORMAT_MAPPING = {
    ContentType.audio_mpeg: AudioFormat.mp3,
    ContentType.audio_json: AudioFormat.json,
    ContentType.audio_pcm: AudioFormat.pcm,
    ContentType.audio_ogg: AudioFormat.ogg_vorbis
}



import functools
import logging
from typing import TYPE_CHECKING, Dict, Set, Optional, Callable

from pydantic import BaseModel, Extra, Any
from pydantic.main import validate_model

from ..utils.case import to_camel

__all__ = ['BasePollyObject']

log = logging.getLogger('aiopolly')


class BasePollyObject(BaseModel):
    class Config:
        arbitrary_types_allowed = True
        extra = Extra.allow
        use_enum_values = True
        validate_assignment = True

    # need to suppress unwanted validation exceptions
    # noinspection PyMissingConstructor
    def __init__(self, **data: Any) -> None:
        if TYPE_CHECKING:  # pragma: no cover
            self.__values__: Dict[str, Any] = {}
            self.__fields_set__: Set[str] = set()
        values, fields_set, error = validate_model(self, data, raise_exc=False)
        object.__setattr__(self, '__values__', values)
        object.__setattr__(self, '__fields_set__', fields_set)

        if error:
            try:
                trust_api_responses = self.polly._trust_api_responses
            except RuntimeError:
                trust_api_responses = True

            if trust_api_responses:
                log.exception('Got unexpected params in API response:', exc_info=error)
            else:
                raise error

    def dict(self, *,
             use_camel_case: bool = False,
             include: Set[str] = None,
             exclude: Set[str] = None,
             by_alias: bool = False,
             skip_defaults: bool = False
             ) -> dict:

        result = BaseModel.dict(self, include=include, exclude=exclude, by_alias=by_alias, skip_defaults=skip_defaults)

        if use_camel_case:
            return to_camel(result)
        return result

    def json(self, *,
             use_camel_case: bool = False,
             include: Set[str] = None,
             exclude: Set[str] = None,
             by_alias: bool = False,
             skip_defaults: bool = False,
             encoder: Optional[Callable[[Any], Any]] = None,
             **dumps_kwargs: Any,
             ) -> str:

        return BaseModel.json(self, use_camel_case=use_camel_case, include=include,
                              exclude=exclude, by_alias=by_alias, skip_defaults=skip_defaults)

    @property
    @functools.lru_cache()
    def polly(self):
        """
        :rtype: aiopolly.polly.Polly
        """
        from .. import Polly
        polly = Polly.get_current()
        if polly is None:
            raise RuntimeError("Can't get polly instance from context. "
                               "You can fix it with setting current instance: "
                               "'Polly.set_current(polly_instance)'")
        return polly

    def __hash__(self):
        def _hash(obj):
            buf = 0
            if isinstance(obj, list):
                for item in obj:
                    buf += _hash(item)
            elif isinstance(obj, dict):
                for dict_key, dict_value in obj.items():
                    buf += hash(dict_key) + _hash(dict_value)
            else:
                try:
                    buf += hash(obj)
                except TypeError:  # Skip unhashable objects
                    pass
            return buf

        result = 0
        for key, value in sorted(self.fields.items()):
            result += hash(key) + _hash(value)

        return result



import functools
import re
from typing import Union, Callable


def string_to_camel(string: str):
    return ''.join(x[:1].upper() + x[1:] for x in string.split('_'))


def string_to_snake(string: str):
    result = re.sub('([A-Z]+)', r'_\1', string).lower()
    if result.startswith('_'):
        return result[1:]
    return result


def to_case(obj: Union[str, dict, list],
            ignore_string=False,
            no_exception=False,
            str_case_converter: Callable = string_to_snake) -> Union[str, dict, list]:
    if isinstance(obj, str):
        return obj if ignore_string else str_case_converter(obj)
    elif isinstance(obj, dict):
        return {str_case_converter(key): to_case(value, ignore_string=True, no_exception=True,
                                                 str_case_converter=str_case_converter)
                for key, value in obj.items()}
    elif isinstance(obj, list):
        return [to_case(element, ignore_string=True, no_exception=True, str_case_converter=str_case_converter)
                for element in obj]
    elif no_exception:
        return obj

    raise ValueError(f'Unexpected type {type(obj)}. Supported types: str, dict, list')


to_snake = functools.partial(to_case, str_case_converter=string_to_snake)
to_camel = functools.partial(to_case, str_case_converter=string_to_camel)
